\ Information
\
\ Parallel Resistor Combination Selector
\
\ Derived from the BASIC program by J.Clerici
\
\ Compile with DX-Forth:  FORTH-F - INCLUDE RESISTOR.F BYE
\

empty forth definitions decimal
application

: TITLE ." RESISTOR version 0.2  19-Apr-10" cr ;

cr .( Compiling: )  title

\ Dependencies
cr .( Requires F.R from FPOUT package and 79-NOT. If you don't )
cr .( have them, enable the equate in source file. ) cr cr

0 [if]

  s" MAX-FLOAT-DIGITS" environment? [if] ( enhanced REPRESENT)
    drop  s" FPOUT34.F" included
  [else]
    s" FPOUT22.F" included
  [then]

: NOT  0= ;

[then]

: HELP  ( -- )
  cr ." Parallel Resistor Combination Selector" cr
  cr ." Given a resistance value and tolerance, this program"
  cr ." will select a parallel combination of resistors to"
  cr ." equal the specified value."
  cr ;

\ define f/p number array
: FARRAY
  create ( n -- )  floats allot
  does> ( n -- a )  swap floats + ;

\ Variables
2 value MODE                    \ 1=E24 2=E12

fvariable RIN                   \ input resistor value
fvariable %IN                   \ input tolerance
fvariable R1                    \ 1st preferred value
fvariable R2                    \ 2nd preferred value
fvariable RP                    \ parallel value

11 farray R1VALUES              \ results arrays
11 farray R2VALUES              \
11 farray RPVALUES              \

create RTABLE  169 floats allot \ table of preferred values
                                \ from 1R to 10M

\ Clear variables and arrays
: CLEARRESULTS  ( -- )
  0.0e
  fdup r1 f!
  fdup r2 f!
  fdup rp f!
  11 0 do
    fdup i r1values f!
    fdup i r2values f!
    fdup i rpvalues f!
  loop
  fdrop ;

clearresults

marker DELETEINIT

\ E24 values for 1 decade
create ESERIES
10 , 11 , 12 , 13 , 15 , 16 , 18 , 20 , 22 , 24 , 27 , 30 ,
33 , 36 , 39 , 43 , 47 , 51 , 56 , 62 , 68 , 75 , 82 , 91 ,

\
: E24  ( u -- ) ( r: -- r )
  cells  eseries +  @  s>d d>f ;

\ Init table with preferred values
: INITTABLE  ( -- )
  24 0 do
    i e24 1e-1 f*  i floats rtable +       f!
    i e24          i floats rtable +  24 floats + f!
    i e24 1e1  f*  i floats rtable +  48 floats + f!
    i e24 1e2  f*  i floats rtable +  72 floats + f!
    i e24 1e3  f*  i floats rtable +  96 floats + f!
    i e24 1e4  f*  i floats rtable + 120 floats + f!
    i e24 1e5  f*  i floats rtable + 144 floats + f!
  loop
  10e6  168 floats rtable +  f! ;

inittable       \ initialise the table
deleteinit      \ delete initialising code

\ Get standard resistor value for index u
: RVAL  ( u -- ) ( r: -- f )
  floats rtable +  f@ ;

\ Calc parallel value of r1 r2
: PARALLEL  ( r: r1 r2 -- r3 )
  fover fover f*  frot frot f+  f/ ;

\ Calc difference of r1 from target value
: DIFF  ( r: r1 -- r2 )
  rin f@  f- ;

\ Calc absolute difference of r1 from target value
: ABSDIFF  ( r: r1 -- r2 )
  diff fabs ;

\ Calc % error
: %ERROR  ( r: r1 -- r2 )
  diff  rin f@  f/  100e f* ;

\ Return true if r is within range
: IN-RANGE?  ( r: r -- ) ( -- fl )
  %error fabs  %in f@  fswap f< 0= ;

\ Store result in slot u
: STOREVALUE  ( u -- )
  >r
  r1 f@  r@ r1values f!
  r2 f@  r@ r2values f!
  rp f@  r> rpvalues f! ;

\ Open space at slot u
: OPENSLOT  ( u -- )
  1-  9 1-
  begin
    2dup > not
  while
    2>r
    r@ r1values  f@  r@ 1+ r1values  f!
    r@ r2values  f@  r@ 1+ r2values  f!
    r@ rpvalues  f@  r@ 1+ rpvalues  f!
    2r> 1-
  repeat
  2drop ;

\ Find a slot
: FINDSLOT  ( -- u ) ( r: f -- )
  9
  begin
    >r
    r@ -1 >
  while
    fdup  r@ rpvalues f@ absdiff f<
  while
    r> 1-
  repeat  then
  fdrop  r> 1+  ( backup 1 place ) ;

\ Add match to results
: ADDTORESULTS  ( r: r1 r2 -- )
  fover r1 f!  fdup  r2 f!  parallel  fdup rp f!
  absdiff
  fdup  9 rpvalues f@ absdiff  f< if ( smaller result )
\    rp f@ f.
    findslot
    dup openslot
    storevalue
  else
    fdrop
  then ;

\ Return true if no more matches are possible
: DONEMATCH?  ( i -- flag )
  >r  rin f@ fdup f+  r> rval  f< ;

\ Find matches
: FINDMATCHES  ( -- )
  169 0 do
    i donematch? if unloop exit then
    169 0 do
      j rval  i rval parallel  in-range? if
        j rval  i rval  addtoresults
      then
    mode +loop
  mode +loop ;

\ Scale resistor and display right-justified
: R.R  ( n1 n2 -- ) ( f: r -- )
  2>r
  fdup 1e3 f< if
    [char] R
  else
    fdup 1e6 f< if
      1e3 f/ [char] K
    else
      1e6 f/ [char] M
    then
  then
  2r> rot >r
  f.r  space  r> emit ;

\ Display results
: SHOWRESULTS  ( -- )
  cr cr
  ."  RESISTOR 1    RESISTOR 2      PARALLEL      ERROR %" cr
  11 0 do
    i rpvalues f@  f0= not if
      cr
      i r1values f@        1  7 r.r
      i r2values f@        1 12 r.r
      i rpvalues f@  fdup  3 14 r.r
      %error               3 11 f.r
    then
  loop
  cr ;

\ Warn if value is already in the E series
: ?IS-PREFERRED  ( -- )
  169 0 do
    i rval  in-range? if
      cr cr ."  This value is in the E"
      mode 1- if ." 12" else ." 24" then ."  series."
      unloop exit
    then
  mode +loop ;

\ Select E12 or E24 series
: SETMODE  ( -- )
  cr ." Select E-Series  (1=E12 2=E24)  ? 1 "  8 emit 8 emit
  key dup [char] 2 = if  emit  1  else  drop  2  then
  to mode  cr ;

\ Get resistor value and tolerance
: GETINPUT  ( -- flag )
  cr ." Input resistor value ? "
  pad  dup 80 accept  >float 0= if  0 exit  then
  fdup f0= if  fdrop 0 exit  then
  rin f!
  cr ." Input tolerance %    ? 1 "  8 emit 8 emit
  pad  dup 80 accept  ?dup if
    >float 0= if 0 exit then
  else
    drop  1e0
  then
  %in f!
  -1 ;

\ Run application
: RESISTOR  ( -- )
  help
  setmode
  begin
    clearresults
    getinput
  while
    ?is-preferred
    findmatches
    showresults
  repeat
  cr ;

\
: DEFAULTS  ( -- )
  2 to mode ;

\ Main
: PROGRAM  ( -- )
  cr title                      \ show program name
  defaults
  resistor ;                    \ execute


cr .( Save to disk? )  y/n
[if]  turnkey program  RESISTOR  [then]
